@@ -52,6 +52,31 @@ enum {
 	MAGNIFIER_CROSSWIRE_COLOR_PROP
 } PropIdx;
 
+static int _x_error = 0;
+
+gboolean
+magnifier_error_check (void)
+{
+	if (_x_error) {
+		_x_error = 0;
+		return TRUE;
+	}
+	return FALSE;
+}
+
+static int
+magnifier_x_error_handler (Display	 *display,
+			   XErrorEvent *error)
+{
+	if (error->error_code == BadAlloc) {
+		_x_error = error->error_code;
+	}
+	else {
+		return -1;
+	}
+	return 0;
+}
+
 static void
 magnifier_realize (GtkWidget *widget)
 {
@@ -92,14 +117,31 @@ magnifier_realize (GtkWidget *widget)
 				 XA_ATOM, 32, PropModeReplace,
 				 (guchar *)wm_type_atoms,
 				 1);
-	
+
+	/* TODO: make sure this works/is reset if the DISPLAY 
+	 * (as well as the SCREEN) changes.
+	 */
+
+	XSetErrorHandler (magnifier_x_error_handler);
 	fprintf (stderr, "Realize complete.\n");
 }
 
+GdkWindow*
+magnifier_get_root (Magnifier *magnifier)
+{
+	if (!magnifier->priv->root) {
+		magnifier->priv->root = gdk_screen_get_root_window (
+			gdk_display_get_screen (magnifier->source_display,
+						magnifier->source_screen_num));
+	}
+	return magnifier->priv->root;
+}
+
 static void
 magnifier_exit (GtkObject *object)
 {
-	gtk_exit (0);
+	gtk_main_quit ();
+	exit (0);
 }
 
 static gint
@@ -196,6 +238,12 @@ magnifier_transform_cursor (Magnifier *magnifier)
 							    width, height);
 		scaled_cursor_pixbuf = gdk_pixbuf_scale_simple (
 			cursor_pixbuf, size, size, GDK_INTERP_NEAREST);
+		
+		magnifier->cursor_hotspot.x = magnifier->priv->cursor_hotspot_x * size 
+					    / magnifier->priv->cursor_default_size;
+		magnifier->cursor_hotspot.y = magnifier->priv->cursor_hotspot_y * size 
+					    / magnifier->priv->cursor_default_size;
+					    
 		scaled_mask_pixbuf = gdk_pixbuf_scale_simple (
 			mask_pixbuf, size, size, GDK_INTERP_NEAREST);
 		g_object_unref (cursor_pixbuf);
@@ -265,7 +313,7 @@ magnifier_init_cursor (Magnifier *magnifier)
 		    fprintf (stderr,"\nHOTSPOT  %d %d\n",
 				magnifier->cursor_hotspot.x,
 				magnifier->cursor_hotspot.y);
-		    gdk_pixbuf_unref (cursor_pixbuf);
+		    g_object_unref (G_OBJECT (cursor_pixbuf));
 		}
 		else
 		    fprintf (stderr,"\nHOTSPOT failed\n");
@@ -287,7 +335,8 @@ magnifier_init_cursor (Magnifier *magnifier)
 				gdk_drawable_get_size (magnifier->priv->cursor,
 						       &magnifier->priv->cursor_default_size,
 						       &magnifier->priv->cursor_default_size);
-
+				magnifier->priv->cursor_hotspot_x = magnifier->cursor_hotspot.x;
+				magnifier->priv->cursor_hotspot_y = magnifier->cursor_hotspot.y;
 				fprintf (stderr, "getting cursor %p, mask %p\n",
 					 magnifier->priv->cursor, magnifier->priv->cursor_mask);
 			}
@@ -699,6 +748,9 @@ magnifier_properties_init (Magnifier *magnifier)
 	bonobo_object_add_interface (BONOBO_OBJECT (magnifier),
 				     BONOBO_OBJECT (magnifier->property_bag));
 
+	def = bonobo_arg_new (BONOBO_ARG_STRING);
+	BONOBO_ARG_SET_STRING (def, getenv ("DISPLAY"));
+
 	bonobo_property_bag_add (magnifier->property_bag,
 				 "source-display-screen",
 				 MAGNIFIER_SOURCE_DISPLAY_PROP,
@@ -715,6 +767,8 @@ magnifier_properties_init (Magnifier *magnifier)
 				 "target display screen",
 				 Bonobo_PROPERTY_WRITEABLE);
 
+	bonobo_arg_release (def);
+
 	bonobo_property_bag_add (magnifier->property_bag,
 				 "source-display-size",
 				 MAGNIFIER_SOURCE_SIZE_PROP,
@@ -752,6 +806,7 @@ magnifier_properties_init (Magnifier *magnifier)
 				 "cursor size, in pixels",
 				 Bonobo_PROPERTY_READABLE |
 				 Bonobo_PROPERTY_WRITEABLE);
+	bonobo_arg_release (def);
 	
 	bonobo_property_bag_add (magnifier->property_bag,
 				 "cursor-scale-factor",
@@ -817,27 +872,8 @@ magnifier_properties_init (Magnifier *magnifier)
 }
 
 static void
-magnifier_init (Magnifier *magnifier)
+magnifier_init_window (Magnifier *magnifier)
 {
-	int ret;
-	
-	magnifier->priv = g_new0 (MagnifierPrivate, 1);
-	magnifier_properties_init (magnifier);
-	magnifier->zoom_regions = NULL;
-	
-	magnifier->source_display = gdk_display_open (getenv ("DISPLAY"));
-	magnifier->target_display = gdk_display_open (getenv ("DISPLAY"));
-	magnifier->source_screen_num = 0;
-	magnifier->target_screen_num = 0;
-	magnifier->cursor_size = 0;
-	magnifier->cursor_scale_factor = 1.0F;
-	magnifier->cursor_color = 0xFF000000;
-	magnifier->crosswire_size = 1;
-	magnifier->crosswire_color = 0;
-	magnifier->crosswire_clip = FALSE;
-	magnifier->cursor_hotspot.x = 0;
-	magnifier->cursor_hotspot.y = 0;
-	magnifier->priv->cursor = NULL;
 	magnifier->priv->w =
 		g_object_connect (gtk_widget_new (gtk_window_get_type (),
 						  "user_data", NULL,
@@ -855,14 +891,34 @@ magnifier_init (Magnifier *magnifier)
 	magnifier->priv->canvas = gtk_fixed_new ();
 	gtk_container_add (GTK_CONTAINER (magnifier->priv->w),
 			   magnifier->priv->canvas);
+	magnifier->priv->root = NULL;
+}
+
+static void
+magnifier_init (Magnifier *magnifier)
+{
+	int ret;
+	
+	magnifier->priv = g_new0 (MagnifierPrivate, 1);
+	magnifier_properties_init (magnifier);
+	magnifier->zoom_regions = NULL;
+	magnifier->source_screen_num = 0;
+	magnifier->target_screen_num = 0;
+	magnifier->cursor_size = 0;
+	magnifier->cursor_scale_factor = 1.0F;
+	magnifier->cursor_color = 0xFF000000;
+	magnifier->crosswire_size = 1;
+	magnifier->crosswire_color = 0;
+	magnifier->crosswire_clip = FALSE;
+	magnifier->cursor_hotspot.x = 0;
+	magnifier->cursor_hotspot.y = 0;
+	magnifier->priv->cursor = NULL;
+	magnifier->priv->w = NULL;
+	magnifier_init_window (magnifier);
 	
 	ret = bonobo_activation_active_server_register (
 		MAGNIFIER_OAFIID, BONOBO_OBJREF (magnifier));
 
-	magnifier->priv->root = gdk_screen_get_root_window (
-		gdk_display_get_screen (magnifier->source_display,
-					magnifier->source_screen_num));
-
 	magnifier->cursor_set = "default";
 }
 
@@ -876,7 +932,7 @@ magnifier_get_cursor (Magnifier *magnifier)
 }
 
 Magnifier *
-magnifier_new (int argc, char **argv)
+magnifier_new (void)
 {
 	return g_object_new (magnifier_get_type(), NULL);
 }
