@@ -54,6 +54,12 @@ enum {
 	ZOOM_REGION_TESTPATTERN_PROP
 } PropIdx;
 
+typedef enum {
+	ZOOM_REGION_ERROR_NONE,
+	ZOOM_REGION_ERROR_NO_TARGET_DRAWABLE,
+	ZOOM_REGION_ERROR_TOO_BIG
+} ZoomRegionPixmapCreationError;
+
 #ifdef TEST_XTST_CURSOR
 static Cursor *x_cursors;
 static Window cursor_window = None;
@@ -492,9 +498,9 @@ zoom_region_clip_to_window (ZoomRegion *zoom_region,
 {
 	GdkRectangle window_rect = {0, 0, 0, 0};
 	if (zoom_region->priv->w->window)
-		gdk_window_get_size (zoom_region->priv->w->window,
-				     &window_rect.x,
-				     &window_rect.y);
+		gdk_drawable_get_size (GDK_DRAWABLE (zoom_region->priv->w->window),
+				       &window_rect.x,
+				       &window_rect.y);
 	return _rectangle_clip_to_rectangle (area, window_rect);
 }
 
@@ -632,7 +638,7 @@ zoom_region_paint_crosswire_cursor (ZoomRegion *zoom_region, GdkRectangle *clip_
 	Magnifier *magnifier = zoom_region->priv->parent;
 	GdkGC *gc;
 	GdkColor color = {0, 0, 0, 0};
-	int x_left_clip, x_right_clip = 0, y_top_clip, y_bottom_clip = 0;
+	int x_left_clip = 0, x_right_clip = 0, y_top_clip = 0, y_bottom_clip = 0;
 	int csize;
 	GdkRectangle rect;
 	GdkDrawable *cursor;
@@ -706,8 +712,8 @@ zoom_region_paint_crosswire_cursor (ZoomRegion *zoom_region, GdkRectangle *clip_
 		}
 		gdk_draw_rectangle (zoom_region->priv->w->window, gc, TRUE,
 				    rect.x - magnifier->crosswire_size / 2,
-				    x_right_clip, magnifier->crosswire_size, 4096);
-		gdk_draw_rectangle (zoom_region->priv->w->window, gc, TRUE, y_bottom_clip,
+				    y_bottom_clip, magnifier->crosswire_size, 4096);
+		gdk_draw_rectangle (zoom_region->priv->w->window, gc, TRUE, x_right_clip,
 				    rect.y - magnifier->crosswire_size / 2,
 				    4096, magnifier->crosswire_size);
 	}
@@ -735,14 +741,16 @@ zoom_region_paint_cursor (ZoomRegion *zoom_region,
 	}
 	if (gdk_rectangle_intersect (clip_rect, &rect, &intersct))
 	{
-		int width, height;
+		int width = 0, height = 0;
 		GdkGC *gc = gdk_gc_new (zoom_region->priv->w->window);
 		GdkDrawable *cursor = magnifier_get_cursor (magnifier);
 		if (!cursor)
 			return;
 		zoom_region->priv->cursor_backing_rect = rect;
-		gdk_drawable_get_size (zoom_region->priv->cursor_backing_pixels,
-				       &width, &height);
+		if (zoom_region->priv->cursor_backing_pixels) {
+			gdk_drawable_get_size (zoom_region->priv->cursor_backing_pixels,
+					       &width, &height);
+		}
 		if (rect.width != width || rect.height != height)
 		{
 			g_object_unref (zoom_region->priv->cursor_backing_pixels);
@@ -752,7 +760,8 @@ zoom_region_paint_cursor (ZoomRegion *zoom_region,
 						rect.height,
 						-1);
 		}
-		gdk_draw_drawable (zoom_region->priv->cursor_backing_pixels,
+		if (zoom_region->priv->w->window != NULL)
+		  gdk_draw_drawable (zoom_region->priv->cursor_backing_pixels,
 				   gc,
 				   zoom_region->priv->w->window,
 				   rect.x,
@@ -851,7 +860,7 @@ zoom_region_paint_pixmap (ZoomRegion *zoom_region,
 	GdkGC *gc = NULL;
 	if (zoom_region->priv->w && zoom_region->priv->w->window)
 		gc = gdk_gc_new (GDK_DRAWABLE (zoom_region->priv->w->window));
-	if (gc)
+	if (gc && zoom_region->priv->pixmap)
 	{
 		gdk_draw_drawable (zoom_region->priv->w->window,
 				   gc,
@@ -862,8 +871,8 @@ zoom_region_paint_pixmap (ZoomRegion *zoom_region,
 				   area->y,
 				   area->width,
 				   area->height);
-		g_object_unref (gc);
 	}
+	g_object_unref (gc); /* OK if gc is NULL */
 }
 
 /**
@@ -883,6 +892,38 @@ zoom_region_paint (ZoomRegion *zoom_region,
 	zoom_region_paint_crosswire_cursor (zoom_region, &paint_area);
 }
 
+static ZoomRegionPixmapCreationError
+zoom_region_create_pixmap (ZoomRegion *zoom_region)
+{
+	if (zoom_region->priv->w && GDK_IS_DRAWABLE (zoom_region->priv->w->window))
+	{
+		long width = (zoom_region->priv->source_area.x2 -
+			      zoom_region->priv->source_area.x1) * zoom_region->xscale;
+		long height = (zoom_region->priv->source_area.y2 -
+			       zoom_region->priv->source_area.y1) * zoom_region->yscale;
+		zoom_region->priv->pixmap =
+			gdk_pixmap_new (
+				zoom_region->priv->w->window,
+				width,
+				height,
+				gdk_drawable_get_depth (
+					zoom_region->priv->w->window));
+
+		if (magnifier_error_check ()) {
+			zoom_region->priv->pixmap = NULL;
+			return ZOOM_REGION_ERROR_TOO_BIG;
+		}
+
+		zoom_region_update (zoom_region,
+				    zoom_region_source_rect_from_view_bounds (
+					    zoom_region,
+					    &zoom_region->priv->exposed_viewport));
+		return ZOOM_REGION_ERROR_NONE;
+	}
+
+	return ZOOM_REGION_ERROR_NO_TARGET_DRAWABLE;
+}
+
 static void
 zoom_region_expose_handler (GtkWindow * w,
 			    GdkEventExpose *event,
@@ -893,23 +934,13 @@ zoom_region_expose_handler (GtkWindow * w,
 
 	if (zoom_region->priv->pixmap == NULL)
 	{
-		if (GDK_IS_DRAWABLE (zoom_region->priv->w->window))
-		{
-			long width = (zoom_region->priv->source_area.x2 -
-				zoom_region->priv->source_area.x1) * zoom_region->xscale;
-			long height = (zoom_region->priv->source_area.y2 -
-				zoom_region->priv->source_area.y1) * zoom_region->yscale;
-			zoom_region->priv->pixmap =
-				gdk_pixmap_new (
-					zoom_region->priv->w->window,
-					width,
-					height,
-					gdk_drawable_get_depth (
-						zoom_region->priv->w->window));
-			zoom_region_update (zoom_region,
-					    zoom_region_source_rect_from_view_bounds (
-						    zoom_region,
-						    &zoom_region->priv->exposed_viewport));
+		/* TODO: scale down if this fails here */
+		while (zoom_region_create_pixmap (zoom_region) ==
+		    ZOOM_REGION_ERROR_TOO_BIG) {
+			zoom_region->xscale -= 1.0;
+			zoom_region->yscale -= 1.0;
+			zoom_region->priv->pixmap = NULL;
+			g_warning ("Scale factor too big to fit in memory; shrinking.");
 		}
 	}
 	zoom_region_paint (zoom_region, &event->area);
@@ -945,9 +976,10 @@ zoom_region_calculate_scroll_rects (ZoomRegion *zoom_region,
 	if (!window)
 		return FALSE;
 
-	gdk_drawable_get_size (GDK_DRAWABLE (window),
-			       &rect.width,
-			       &rect.height);
+	if (window != NULL)
+	  gdk_drawable_get_size (GDK_DRAWABLE (window),
+				 &rect.width,
+				 &rect.height);
 
 	if ((ABS (dx) >= rect.width) || (ABS (dy) >= rect.height)) {
 		*scroll_rect = rect;
@@ -1110,7 +1142,7 @@ zoom_region_update_pointer (ZoomRegion *zoom_region, gboolean draw_cursor)
 	if (!zoom_region->priv->parent) return FALSE;
 	magnifier = zoom_region->priv->parent;
 	/* TODO: there's really no reason we should be using magnifier->priv->root here */
-	if (magnifier && magnifier->priv && magnifier->priv->root)
+	if (magnifier && magnifier->priv && magnifier_get_root (magnifier))
 	{
 #ifdef TEST_XTST_CURSOR
 		int max_cursors = XC_num_glyphs;
@@ -1174,7 +1206,7 @@ zoom_region_update_pointer (ZoomRegion *zoom_region, gboolean draw_cursor)
 		}
 #endif
 		gdk_window_get_pointer (
-			magnifier->priv->root,
+			magnifier_get_root (magnifier),
 			&mouse_x_return,
 			&mouse_y_return,
 			&mask_return);
@@ -1423,7 +1455,6 @@ zoom_region_update (ZoomRegion *zoom_region,
 
 		zoom_region_post_process_pixbuf (zoom_region, subimage,
 						 zoom_region->priv->scaled_pixbuf);
-		
 		gc = gdk_gc_new (zoom_region->priv->w->window);
 
 		gdk_pixbuf_render_to_drawable (zoom_region->priv->scaled_pixbuf,
@@ -1438,6 +1469,8 @@ zoom_region_update (ZoomRegion *zoom_region,
 					       GDK_RGB_DITHER_NONE,
 					       0,
 					       0);
+		if (magnifier_error_check ())
+			g_warning ("Could not render scaled image to drawable; out of memory!\n");
 		g_object_unref (gc);
 		g_object_unref (subimage);
 		gdk_window_begin_paint_rect (zoom_region->priv->w->window, &paint_rect);
@@ -1461,11 +1494,11 @@ zoom_region_init_window (ZoomRegion *zoom_region)
 	border = gtk_fixed_new ();
 	zoom_region->priv->border = border;
 	zoom_region->priv->w = zoomer;
-	gtk_drawing_area_size (GTK_DRAWING_AREA (zoomer),
-			       zoom_region->viewport.x2 -
-			       zoom_region->viewport.x1 - zoom_region->border_size * 2,
-			       zoom_region->viewport.y2 -
-			       zoom_region->viewport.y1 - zoom_region->border_size * 2);
+	gtk_widget_set_size_request (GTK_WIDGET (zoomer),
+				     zoom_region->viewport.x2 -
+				     zoom_region->viewport.x1 - zoom_region->border_size * 2,
+				     zoom_region->viewport.y2 -
+				     zoom_region->viewport.y1 - zoom_region->border_size * 2);
 	fprintf (stderr, "putting into fixed\n");
 	gtk_fixed_put (parent, border,
 		       zoom_region->viewport.x1,
@@ -1478,9 +1511,9 @@ zoom_region_init_window (ZoomRegion *zoom_region)
 	gtk_widget_show (GTK_WIDGET (parent));
 	fprintf (stderr, "shown\n");
 	zoom_region->priv->expose_handler_id =
-		gtk_signal_connect (GTK_OBJECT (zoom_region->priv->w),
+		g_signal_connect (G_OBJECT (zoom_region->priv->w),
 			    "expose_event",
-			    GTK_SIGNAL_FUNC (zoom_region_expose_handler),
+			    G_CALLBACK (zoom_region_expose_handler),
 			    zoom_region);
 	fprintf (stderr, "New window created\n");
 	gtk_widget_show (GTK_WIDGET (zoom_region->priv->w));
@@ -1575,11 +1608,11 @@ zoom_region_set_viewport (ZoomRegion *zoom_region,
 				zoom_region->priv->border,
 				zoom_region->viewport.x1,
 				zoom_region->viewport.y1);
-		gtk_drawing_area_size (GTK_DRAWING_AREA (zoom_region->priv->w),
-				       zoom_region->viewport.x2 -
-				       zoom_region->viewport.x1,
-				       zoom_region->viewport.y2 -
-				       zoom_region->viewport.y1);
+		gtk_widget_set_size_request (GTK_WIDGET (zoom_region->priv->w),
+					     zoom_region->viewport.x2 -
+					     zoom_region->viewport.x1,
+					     zoom_region->viewport.y2 -
+					     zoom_region->viewport.y1);
 	}
 	zoom_region_queue_update (zoom_region,
 				  zoom_region_source_rect_from_view_bounds (
@@ -1738,7 +1771,7 @@ impl_zoom_region_set_roi (PortableServer_Servant servant,
 	zoom_region_align (zoom_region);
 }
 
-static void
+static CORBA_boolean
 impl_zoom_region_set_mag_factor (PortableServer_Servant servant,
 				 const CORBA_float mag_factor_x,
 				 const CORBA_float mag_factor_y,
@@ -1747,6 +1780,10 @@ impl_zoom_region_set_mag_factor (PortableServer_Servant servant,
 	ZoomRegion *zoom_region =
 		ZOOM_REGION (bonobo_object_from_servant (servant));
 	CORBA_any *any;
+	double xs_old = zoom_region->xscale;
+	double ys_old = zoom_region->yscale;
+	CORBA_boolean retval = CORBA_TRUE;
+
 	/* TODO: assert that parent is magnifier object */
 	Bonobo_PropertyBag properties =
 		GNOME_Magnifier_Magnifier_getProperties(
@@ -1757,6 +1794,9 @@ impl_zoom_region_set_mag_factor (PortableServer_Servant servant,
 	if (!BONOBO_EX (ev))
 		zoom_region->priv->source_area =
 			*((GNOME_Magnifier_RectBounds *) any->_value);
+	else
+		retval = CORBA_FALSE;
+
 	zoom_region->xscale = mag_factor_x;
 	zoom_region->yscale = mag_factor_y;
 
@@ -1774,10 +1814,25 @@ impl_zoom_region_set_mag_factor (PortableServer_Servant servant,
 		g_object_unref (zoom_region->priv->pixmap);
 		zoom_region->priv->pixmap = NULL;
 	}
+	if (zoom_region_create_pixmap (zoom_region) == ZOOM_REGION_ERROR_TOO_BIG) {
+		zoom_region->xscale = xs_old;
+		zoom_region->yscale = ys_old;
+		zoom_region_create_pixmap (zoom_region);
+		g_object_unref (zoom_region->priv->scaled_pixbuf);
+
+		zoom_region->priv->scaled_pixbuf = gdk_pixbuf_new (
+			GDK_COLORSPACE_RGB, FALSE, 8,
+			zoom_region->priv->source_area.x2 -
+			zoom_region->priv->source_area.x1 * zoom_region->xscale + 1,
+			zoom_region->priv->source_area.y2 -
+			zoom_region->priv->source_area.y1 * zoom_region->yscale + 1);
+		retval = CORBA_FALSE;
+	}
 	zoom_region_update_current (zoom_region);
 	zoom_region_sync (zoom_region);
 
 	bonobo_object_release_unref (properties, NULL);
+	return retval;
 }
 
 static void
@@ -1910,7 +1965,8 @@ zoom_region_properties_init (ZoomRegion *zoom_region)
 				 "If false, zoom region does not auto-update, but is drawn into directly by the client",
 				 Bonobo_PROPERTY_READABLE |
 				 Bonobo_PROPERTY_WRITEABLE);
-	
+
+	bonobo_arg_release (def);
 	def = bonobo_arg_new (BONOBO_ARG_INT);
 	BONOBO_ARG_SET_INT (def, GNOME_Magnifier_ZoomRegion_SCROLL_FASTEST);
 	
@@ -1923,6 +1979,7 @@ zoom_region_properties_init (ZoomRegion *zoom_region)
 				 Bonobo_PROPERTY_READABLE |
 				 Bonobo_PROPERTY_WRITEABLE);
 
+	bonobo_arg_release (def);
 	def = bonobo_arg_new (BONOBO_ARG_BOOLEAN);
 	BONOBO_ARG_SET_BOOLEAN (def, FALSE);
 
@@ -1935,6 +1992,7 @@ zoom_region_properties_init (ZoomRegion *zoom_region)
 				 Bonobo_PROPERTY_READABLE |
 				 Bonobo_PROPERTY_WRITEABLE);
 
+	bonobo_arg_release (def);
 	def = bonobo_arg_new (BONOBO_ARG_BOOLEAN);
 	BONOBO_ARG_SET_BOOLEAN (def, TRUE);
 	
@@ -1947,7 +2005,7 @@ zoom_region_properties_init (ZoomRegion *zoom_region)
 				 Bonobo_PROPERTY_READABLE |
 				 Bonobo_PROPERTY_WRITEABLE);
 
-	def = bonobo_arg_new (BONOBO_ARG_STRING);
+	bonobo_arg_release (def);
 
 	bonobo_property_bag_add (zoom_region->properties,
 				 "smoothing-type",
@@ -1970,6 +2028,7 @@ zoom_region_properties_init (ZoomRegion *zoom_region)
 				 Bonobo_PROPERTY_READABLE |
 				 Bonobo_PROPERTY_WRITEABLE);
 
+	bonobo_arg_release (def);
 	def = bonobo_arg_new (BONOBO_ARG_FLOAT);
 	BONOBO_ARG_SET_FLOAT (def, 2.0);
 
@@ -1982,6 +2041,7 @@ zoom_region_properties_init (ZoomRegion *zoom_region)
 				 Bonobo_PROPERTY_READABLE |
 				 Bonobo_PROPERTY_WRITEABLE);
 
+	bonobo_arg_release (def);
 	def = bonobo_arg_new (BONOBO_ARG_FLOAT);
 	BONOBO_ARG_SET_FLOAT (def, 2.0);
 
@@ -1994,6 +2054,7 @@ zoom_region_properties_init (ZoomRegion *zoom_region)
 				 Bonobo_PROPERTY_READABLE |
 				 Bonobo_PROPERTY_WRITEABLE);
 
+	bonobo_arg_release (def);
 	def = bonobo_arg_new (BONOBO_ARG_INT);
 	BONOBO_ARG_SET_INT (def, 0);
 	
@@ -2006,6 +2067,7 @@ zoom_region_properties_init (ZoomRegion *zoom_region)
 				 Bonobo_PROPERTY_READABLE |
 				 Bonobo_PROPERTY_WRITEABLE);
 
+	bonobo_arg_release (def);
 	def = bonobo_arg_new (BONOBO_ARG_LONG);
 	BONOBO_ARG_SET_LONG (def, 0x00000000);
 	
@@ -2018,6 +2080,7 @@ zoom_region_properties_init (ZoomRegion *zoom_region)
 				 Bonobo_PROPERTY_READABLE |
 				 Bonobo_PROPERTY_WRITEABLE);
 
+	bonobo_arg_release (def);
 	def = bonobo_arg_new (BONOBO_ARG_INT);
 	BONOBO_ARG_SET_INT (def, 0);
 
@@ -2030,6 +2093,7 @@ zoom_region_properties_init (ZoomRegion *zoom_region)
 				 Bonobo_PROPERTY_READABLE |
 				 Bonobo_PROPERTY_WRITEABLE);
 
+	bonobo_arg_release (def);
 	def = bonobo_arg_new (BONOBO_ARG_INT);
 	BONOBO_ARG_SET_INT (def, 0);
 
@@ -2041,6 +2105,7 @@ zoom_region_properties_init (ZoomRegion *zoom_region)
 				 "y-alignment policy for this region",
 				 Bonobo_PROPERTY_READABLE |
 				 Bonobo_PROPERTY_WRITEABLE);
+	bonobo_arg_release (def);
 
 	bonobo_property_bag_add (zoom_region->properties,
 				 "viewport",
